import {DependencyEntry} from "./dependency-entry";
import { EventEmitter } from "events";


export interface DependencyManager {

  /**
   * Add the dependency. If the dependency already exists it will replace it
   * @param name The dependency name
   * @param dependency The dependency
   * @throws {Error} If the argument are invalid
   */
  add(name: string, dependency: any): void;

  /**
   * Remove the dependency
   * @param  name The dependency name
   */
  remove(name: string): void;

  /**
   * Check if the dependency manager contains the dependency
   * @param name The dependency name
   * @return true if exists, otherwise false
   */
  contains(name: string): boolean;

  /**
   * Retrieve the dependency
   * @param name The dependency name
   * @param defaultValue The default dependency to return if not found
   */
  get(name: string, defaultValue?: any): any;

  /**
   * Returns the dependency names
   */
  dependencyNames(): string[];
}

export namespace simpleDependencyManager {

  export interface ContainsAllResponseItem {
    /**
     * The dependency names
     */
    names: string[];
  }

  interface ContainsAllResponse {
    /**
     * The dependencies found
     */
    found: ContainsAllResponseItem;
    /**
     * The dependencies not found
     */
    notFound: ContainsAllResponseItem;
  }

  export interface DependencyIndexes {
    byName: object;
  }

  export interface Properties {
    /**
     * Contains all dependencies
     */
    dependencies: any[];

    /**
     * Contains the indexes
     */
    dependencyIndexes: DependencyIndexes;
  }
}

/**
 * Implementation of the dependency manager
 */
export class SimpleDependencyManager extends EventEmitter implements DependencyManager {

  /**
   * Internal properties
   */
  protected properties: simpleDependencyManager.Properties;

  /**
   * Returns the signal names
   *
   * Same specification of the base-object
   */
  public static getSignalNames(): string[];

  /**
   * Find the index of the entry by the dependency name
   * @param name The dependency name
   * @return The index of the entry or -1 if not found
   */
  public findIndexByName(name: string): Number;

  /**
   * Try to find the dependency
   * @param {*} arg The argument to find the index
   * @return The index if found otherwise -1
   */
  public indexOf(arg: string| DependencyEntry): Number;

  /**
   * Format the name
   * @param name The name to format
   * @return The name formatted
   * @throw {TypeError} If the name is invalid
   */
  public formatDependencyName(name: string): string;

  /**
   * Test if the name is valid
   * @param name The name to test
   * @return true if valid, otherwise false
   */
  public isValidName(name: string): boolean;

  /**
   * Replace the dependency by the new one
   * @param old The dependency name
   * @param name The new name
   * @param dependency The dependency
   * @throws {Error} If the argument are invalid
   */
  public replace(old: string, name: string, dependency: any): void;

  /**
   * @inheritDoc
   */
  public add(name: string, dependency: any): void;

  /**
   * @inheritDoc
   */
  public remove(name: string): void;

  /**
   * @inheritDoc
   */
  public contains(name: string): boolean;

  /**
   * Check if the dependencies exist
   * @param names The dependency names
   * @return the result of the check
   */
  public containAll(names: string | string[]): simpleDependencyManager.ContainsAllResponse;

  /**
   * @inheritDoc
   */
  public get(name: string, defaultValue?: any): any;

  /**
   * @inheritDoc
   */
  public dependencyNames(): string[];

  /**
   * Returns all dependencies
   */
  public all(): any[];

  /**
   * Remove all the indexes
   */
  protected _removeIndexes();

  /**
   * Refresh all indexes
   */
  protected _refreshIndexes();
}
