const _ = require('lodash'),
  EventEmitter = require('events').EventEmitter,
  SimpleDependencyEntry = require('./dependency-entry').SimpleDependencyEntry;

const internalSignalNames = ['dependencyRemoved', 'dependencyAdded', 'dependencyReplaced'];

/**
 * Simple dependency manager
 */
class SimpleDependencyManager extends EventEmitter {

  constructor() {
    super();

    Object.defineProperty(this, 'properties', {
      enumerable: false,
      writable: true,
      value: _.assign({}, {
        dependencies: [],
        dependencyIndexes: {
          byName: {}
        }
      })
    });
  }

  static getSignalNames() {
    return internalSignalNames;
  }

  _refreshIndexes() {
    // Clear dependencies
    this._removeIndexes();
    _.each(this.properties.dependencies, (dependency, index) => {
      this._registerIndex(dependency, index);
    });
  }

  _removeIndexes() {
    this.properties.dependencyIndexes = {
      byName: {}
    };
  }

  _removeIndex(entry, index) {
    if (!(entry instanceof SimpleDependencyEntry)) {
      throw new TypeError("Invalid entry type");
    } else if (!_.isNumber(index)) {
      throw new TypeError(`Invalid type '${typeof index}' for index`);
    }

    // Remove the property
    delete this.properties.dependencyIndexes.byName[entry.name];
  }

  _registerIndex(entry, index) {
    if (!(entry instanceof SimpleDependencyEntry)) {
      throw new TypeError("Invalid entry type");
    } else if (!_.isNumber(index)) {
      throw new TypeError(`Invalid type '${typeof index}' for index`);
    }

    this.properties.dependencyIndexes.byName[entry.name] = index;
  }

  findIndexByName(name) {
    if (!this.isValidName(name)) {
      return -1;
    }

    name = this.formatDependencyName(name);
    return _.isNumber(this.properties.dependencyIndexes.byName[name]) ? this.properties.dependencyIndexes.byName[name] : -1;
  }

  indexOf(arg) {
    if (this.isValidName(arg)) {
      return this.findIndexByName(arg);
    } else if (arg instanceof SimpleDependencyEntry) {
      return this.findIndexByName(arg.name);
    }

    return -1;
  }

  formatDependencyName(name) {
    return SimpleDependencyEntry.formatName(name);
  }

  isValidName(name) {
    return SimpleDependencyEntry.isValidName(name);
  }

  add(name, dependency) {
    if (!this.isValidName(name)) {
      throw new TypeError("Invalid dependency name");
    }

    const entry = SimpleDependencyEntry.of(name, dependency);
    const currentIndex = this.indexOf(name);
    if (currentIndex >= 0) {
      // remove
      const oldDependency = this.properties.dependencies[currentIndex].dependency;
      this.properties.dependencies[currentIndex] = entry;
      this.emit("dependencyReplaced", entry.name, dependency, oldDependency);
    } else {
      const newIndex = this.properties.dependencies.length;
      this.properties.dependencies.push(entry);
      this._registerIndex(entry, newIndex);
      this.emit("dependencyAdded", entry.name, entry.dependency, entry);
    }
  }

  replace(old, name, dependency) {
    if (!this.isValidName(old)) {
      throw new TypeError("Invalid old dependency name");
    } else if (!this.isValidName(name)) {
      throw new TypeError("Invalid dependency name");
    }

    // Remove the old dependency
    this.remove(old);
    this.add(name, dependency);
  }

  remove(name) {
    const index = this.indexOf(name);
    if (index >= 0) {
      const entry = this.properties.dependencies[index];
      _.remove(this.properties.dependencies, (value, elementIndex) => index === elementIndex);
      this.emit("dependencyRemoved", entry.name, entry.dependency, entry);
      this._refreshIndexes();
    }
  }

  contains(name) {
    return this.indexOf(name) >= 0;
  }

  containAll(names) {
    const response = {
      found: {
        names: []
      },
      notFound: {
        names: []
      }
    };

    if (!_.isArray(names)) {
      names = [names];
    }

    _.each(names, name => {
      if (this.contains(name)) {
        response.found.names.push(name);
      } else {
        response.notFound.names.push(name);
      }
    });

    return response;
  }

  get(name, defaultValue = undefined) {
    const index = this.indexOf(name);
    return index >= 0 ? this.properties.dependencies[index].dependency : defaultValue;
  }

  dependencyNames() {
    return _.sortBy(_.keys(this.properties.dependencyIndexes.byName));
  }

  all() {
    return _.concat([], this.properties.dependencies);
  }
}

exports.SimpleDependencyManager = module.exports.SimpleDependencyManager = SimpleDependencyManager;
