const _ = require('lodash'),
  EventEmitter = require('events').EventEmitter;

class SimpleMultiDependencyManager extends EventEmitter {

  constructor(primary, secondaries) {
    super();

    if (!SimpleMultiDependencyManager.isValidManager(primary)) {
      throw new TypeError(`Invalid primary manager`);
    } else if (!_.isNil(secondaries) && !SimpleMultiDependencyManager.isValidSecondaries(secondaries)) {
      throw new TypeError(`Invalid secondaries`);
    }

    Object.defineProperty(this, 'properties', {
      enumerable: false,
      writable: true,
      value: _.assign({}, {
        primary: primary,
        secondaries: _.concat([], secondaries || [])
      })
    });
  }

  static isValidManager(manager) {
    return _.isObjectLike(manager) &&
      _.isFunction(manager.add) && _.isFunction(manager.get) &&
      _.isFunction(manager.remove) && _.isFunction(manager.contains);
  }

  static isValidSecondaries(secondaries) {
    if (!_.isArray(secondaries)) {
      return false;
    }

    let valid = true;
    _.each(secondaries, item => {
      valid = SimpleMultiDependencyManager.isValidManager(item);
      if (!valid) {
        return false;
      }
    });

    return valid;
  }

  get primary() {
    return this.properties.primary;
  }

  get secondaries() {
    return _.concat([], this.properties.secondaries);
  }

  addSecondary(manager) {
    if (_.isNil(manager)) {
      return;
    } else if (SimpleMultiDependencyManager.isValidManager(manager)) {
      this.properties.secondaries.push(manager);
    } else {
      throw new TypeError(`Invalid input manager`);
    }
  }

  contains(name) {
    if (this.properties.primary.contains(name)) {
      return true;
    }

    let found = false;
    _.each(this.properties.secondaries, manager => {
      if (manager.contains(name)) {
        found = true;
        return false;
      }
    });
    return found;
  }

  get(name, defaultValue = undefined) {
    if (this.properties.primary.contains(name)) {
      return this.properties.primary.get(name);
    }

    let found = defaultValue;
    _.each(this.properties.secondaries, manager => {
      if (manager.contains(name)) {
        found = manager.get(name);
        return false;
      }
    });
    return found;
  }

  add(name, value) {
    return this.properties.primary.add(name, value);
  }

  remove(name) {
    return this.properties.primary.remove(name);
  }

  dependencyNames() {
    let names = _.concat([], this.properties.primary.dependencyNames());

    _.each(this.properties.secondaries, manager => {
      names = _.concat(names, manager.dependencyNames());
    });

    return _.sortBy(_.uniq(names));
  }
}

exports.SimpleMultiDependencyManager = module.exports.SimpleMultiDependencyManager = SimpleMultiDependencyManager;
