describe('Unit Test - SimpleMultiDependencyManager', function () {
  const SimpleMultiDependencyManager = require('../../../lib/multi-dependency-manager').SimpleMultiDependencyManager,
    _ = require("lodash");

  /**
   * Create an instance
   * @return {SimpleMultiDependencyManager} The instance
   */
  function createInstance() {
    spyOn(SimpleMultiDependencyManager, 'isValidManager').and.returnValue(true);
    spyOn(SimpleMultiDependencyManager, 'isValidSecondaries').and.returnValue(true);

    return new SimpleMultiDependencyManager();
  }

  describe('#construct', function () {
    it('Given invalid primary Then must throw error', function () {
      spyOn(SimpleMultiDependencyManager, 'isValidManager').and.returnValue(false);
      spyOn(SimpleMultiDependencyManager, 'isValidSecondaries').and.returnValue(false);

      expect(() => {
        new SimpleMultiDependencyManager('a');
      }).toThrow(jasmine.any(TypeError));
      expect(SimpleMultiDependencyManager.isValidManager).toHaveBeenCalledTimes(1);
      expect(SimpleMultiDependencyManager.isValidManager).toHaveBeenCalledWith('a');

      expect(SimpleMultiDependencyManager.isValidSecondaries).not.toHaveBeenCalled();
    });

    it('Given invalid secondaries Then must throw error', function () {
      spyOn(SimpleMultiDependencyManager, 'isValidManager').and.returnValue(true);
      spyOn(SimpleMultiDependencyManager, 'isValidSecondaries').and.returnValue(false);

      expect(() => {
        new SimpleMultiDependencyManager('a', 'c');
      }).toThrow(jasmine.any(TypeError));
      expect(SimpleMultiDependencyManager.isValidManager).toHaveBeenCalledTimes(1);
      expect(SimpleMultiDependencyManager.isValidManager).toHaveBeenCalledWith('a');

      expect(SimpleMultiDependencyManager.isValidSecondaries).toHaveBeenCalledTimes(1);
      expect(SimpleMultiDependencyManager.isValidSecondaries).toHaveBeenCalledWith('c');
    });

    it('Given valid primary and secondary undefined Then must return expected instance', function () {
      spyOn(SimpleMultiDependencyManager, 'isValidManager').and.returnValue(true);
      spyOn(SimpleMultiDependencyManager, 'isValidSecondaries').and.returnValue(true);

      const instance = new SimpleMultiDependencyManager('a', undefined);
      expect(instance.properties.primary).toBe('a');
      expect(instance.properties.secondaries).toEqual([]);

      expect(SimpleMultiDependencyManager.isValidManager).toHaveBeenCalledTimes(1);
      expect(SimpleMultiDependencyManager.isValidManager).toHaveBeenCalledWith('a');

      expect(SimpleMultiDependencyManager.isValidSecondaries).not.toHaveBeenCalled();
    });

    it('Given valid primary and secondary null Then must return expected instance', function () {
      spyOn(SimpleMultiDependencyManager, 'isValidManager').and.returnValue(true);
      spyOn(SimpleMultiDependencyManager, 'isValidSecondaries').and.returnValue(true);

      const instance = new SimpleMultiDependencyManager('a', null);
      expect(instance.properties.primary).toBe('a');
      expect(instance.properties.secondaries).toEqual([]);

      expect(SimpleMultiDependencyManager.isValidManager).toHaveBeenCalledTimes(1);
      expect(SimpleMultiDependencyManager.isValidManager).toHaveBeenCalledWith('a');

      expect(SimpleMultiDependencyManager.isValidSecondaries).not.toHaveBeenCalled();
    });

    it('Given valid primary and secondaries Then must return expected instance', function () {
      spyOn(SimpleMultiDependencyManager, 'isValidManager').and.returnValue(true);
      spyOn(SimpleMultiDependencyManager, 'isValidSecondaries').and.returnValue(true);

      const instance = new SimpleMultiDependencyManager('a', ['c', 'd']);
      expect(instance.properties.primary).toBe('a');
      expect(instance.properties.secondaries).toEqual(['c', 'd']);

      expect(SimpleMultiDependencyManager.isValidManager).toHaveBeenCalledTimes(1);
      expect(SimpleMultiDependencyManager.isValidManager).toHaveBeenCalledWith('a');

      expect(SimpleMultiDependencyManager.isValidSecondaries).toHaveBeenCalledTimes(1);
      expect(SimpleMultiDependencyManager.isValidSecondaries).toHaveBeenCalledWith(['c', 'd']);
    });
  }); // #construct

  describe("When require the manager", function () {
    it("Must be a function", function () {
      expect(SimpleMultiDependencyManager).toEqual(jasmine.any(Function));
    });
  }); // End when require

  describe('#isValidSecondaries', function () {
    it('Given no array Then must return false', function () {
      const invalidManagers = [
        undefined, null, {}, 1, false
      ];
      _.each(invalidManagers, item => {
        expect(SimpleMultiDependencyManager.isValidSecondaries(item)).toBeFalsy();
      });
    });

    it('Given invalid manager in array Then must return false', function () {
      spyOn(SimpleMultiDependencyManager, 'isValidManager').and.returnValues(true, false);

      const secondaries = [{
        add: () => {},
        get: () => {},
        remove: () => {},
        contains: () => {}
      }, {}, {}];
      expect(SimpleMultiDependencyManager.isValidSecondaries(secondaries)).toBeFalsy();

      expect(SimpleMultiDependencyManager.isValidManager).toHaveBeenCalledTimes(2);
      expect(SimpleMultiDependencyManager.isValidManager).toHaveBeenCalledWith(secondaries[0]);
      expect(SimpleMultiDependencyManager.isValidManager).toHaveBeenCalledWith(secondaries[1]);
    });

    it('Given invalid manager in array Then must return false', function () {
      spyOn(SimpleMultiDependencyManager, 'isValidManager').and.returnValue(true);

      const secondaries = [];
      for (let i = 0; i < 3; ++i) {
        secondaries.push({
          add: () => {},
          get: () => {},
          remove: () => {},
          contains: () => {}
        });
      }
      expect(SimpleMultiDependencyManager.isValidSecondaries(secondaries)).toBeTruthy();

      expect(SimpleMultiDependencyManager.isValidManager).toHaveBeenCalledTimes(3);
      _.each(secondaries, item => {
        expect(SimpleMultiDependencyManager.isValidManager).toHaveBeenCalledWith(item);
      });
    });
  }); // #isValidSecondaries

  describe('#isValidManager', function () {
    it("Given invalid manager Then must return false", function () {
      const invalidManagers = [
        undefined, null, [], {}, 1, false
      ];
      _.each(invalidManagers, item => {
        expect(SimpleMultiDependencyManager.isValidManager(item)).toBeFalsy();
      });
    });

    it('Given valid manager Then must return true', function () {
      expect(SimpleMultiDependencyManager.isValidManager({
        add: () => {},
        get: () => {},
        remove: () => {},
        contains: () => {}
      })).toBeTruthy();
    });
  }); // #isValidManager

  describe('#primary', function () {
    it("Then must return expected response", function () {
      const instance = createInstance();

      instance.properties.primary = {
        a: 12
      };
      expect(instance.primary).toBe(instance.properties.primary);
    });
  });

  describe('#secondaries', function () {
    it('Then must return expected response', function () {
      const instance = createInstance();

      instance.properties.secondaries = ['a', 'b'];
      const secondaries = instance.secondaries;
      expect(secondaries).toEqual(instance.properties.secondaries);
      expect(secondaries).not.toBe(instance.properties.secondaries);
    });
  }); // #secondaries

  describe('#addSecondary', function () {
    it("Given null Then must do noting", function () {
      const instance = createInstance();

      instance.properties.secondaries = ['a', 'b'];

      instance.addSecondary(undefined);
      expect(instance.properties.secondaries).toEqual(['a', 'b']);

      instance.addSecondary(null);
      expect(instance.properties.secondaries).toEqual(['a', 'b']);
    });

    it('Given valid manager Then must add it', function () {
      const instance = createInstance();

      const manager = {
        add: () => {},
        get: () => {},
        remove: () => {},
        contains: () => {}
      };

      instance.properties.secondaries = ['a', 'b'];
      SimpleMultiDependencyManager.isValidManager.and.returnValue(true);

      instance.addSecondary(manager);
      expect(instance.properties.secondaries).toEqual(['a', 'b', manager]);
      expect(SimpleMultiDependencyManager.isValidManager).toHaveBeenCalledWith(manager);
    });

    it('Given invalid manager Then must throw error', function () {
      const instance = createInstance();

      SimpleMultiDependencyManager.isValidManager.and.returnValue(false);

      expect(() => {
        instance.addSecondary({});
      }).toThrow(jasmine.any(TypeError));
    });
  }); // #addSecondary

  describe('#contains', function () {
    it('Given primary and contains returns true Then must return true', function () {
      const instance = createInstance();

      instance.properties.primary = {
        contains: jasmine.createSpy('contains').and.returnValue(true)
      };

      instance.properties.secondaries = [{
        contains: jasmine.createSpy('contains').and.returnValue(false)
      }];

      expect(instance.contains('a')).toBeTruthy();
      expect(instance.properties.primary.contains).toHaveBeenCalledTimes(1);
      expect(instance.properties.primary.contains).toHaveBeenCalledWith('a');

      expect(instance.properties.secondaries[0].contains).not.toHaveBeenCalled();
    });

    it('Given only secondary returns false Then must return true', function () {
      const instance = createInstance();

      instance.properties.primary = {
        contains: jasmine.createSpy('contains').and.returnValue(false)
      };

      instance.properties.secondaries = [{
        contains: jasmine.createSpy('contains').and.returnValue(false)
      }, {
        contains: jasmine.createSpy('contains').and.returnValue(true)
      }, {
        contains: jasmine.createSpy('contains').and.returnValue(false)
      }];

      expect(instance.contains('a')).toBeTruthy();
      expect(instance.properties.primary.contains).toHaveBeenCalledTimes(1);
      expect(instance.properties.primary.contains).toHaveBeenCalledWith('a');

      expect(instance.properties.secondaries[0].contains).toHaveBeenCalledTimes(1);
      expect(instance.properties.secondaries[0].contains).toHaveBeenCalledWith('a');

      expect(instance.properties.secondaries[1].contains).toHaveBeenCalledTimes(1);
      expect(instance.properties.secondaries[1].contains).toHaveBeenCalledWith('a');

      expect(instance.properties.secondaries[2].contains).not.toHaveBeenCalled();
    });

    it('Given all contains returns false Then must return false', function () {
      const instance = createInstance();

      instance.properties.primary = {
        contains: jasmine.createSpy('contains').and.returnValue(false)
      };

      instance.properties.secondaries = [{
        contains: jasmine.createSpy('contains').and.returnValue(false)
      }, {
        contains: jasmine.createSpy('contains').and.returnValue(false)
      }, {
        contains: jasmine.createSpy('contains').and.returnValue(false)
      }];

      expect(instance.contains('a')).toBeFalsy();
      expect(instance.properties.primary.contains).toHaveBeenCalledTimes(1);
      expect(instance.properties.primary.contains).toHaveBeenCalledWith('a');

      _.each(instance.properties.secondaries, manager => {
        expect(manager.contains).toHaveBeenCalledTimes(1);
        expect(manager.contains).toHaveBeenCalledWith('a');
      });
    });
  }); // #contains

  describe('#get', function () {
    it('Given primary to return manager Then must return expected response', function () {
      const instance = createInstance();

      const expectedResponse = {
        a: 9
      };
      instance.properties.primary = {
        contains: jasmine.createSpy('contains').and.returnValue(true),
        get: jasmine.createSpy('get').and.returnValue(expectedResponse)
      };

      expect(instance.get('c')).toBe(expectedResponse);
      expect(instance.properties.primary.contains).toHaveBeenCalledTimes(1);
      expect(instance.properties.primary.contains).toHaveBeenCalledWith('c');
      expect(instance.properties.primary.get).toHaveBeenCalledTimes(1);
      expect(instance.properties.primary.get).toHaveBeenCalledWith('c');
    });

    it('Given no dependency Then must return expected response', function () {
      const instance = createInstance();

      const expectedResponse = {
        a: 9
      };
      instance.properties.primary = {
        contains: jasmine.createSpy('contains').and.returnValue(false),
        get: jasmine.createSpy('get')
      };
      instance.properties.secondaries = [{
        contains: jasmine.createSpy('contains').and.returnValue(false)
      }, {
        contains: jasmine.createSpy('contains').and.returnValue(false)
      }];

      expect(instance.get('c', expectedResponse)).toBe(expectedResponse);
      expect(instance.properties.primary.contains).toHaveBeenCalledTimes(1);
      expect(instance.properties.primary.contains).toHaveBeenCalledWith('c');

      _.each(instance.properties.secondaries, item => {
        expect(item.contains).toHaveBeenCalledTimes(1);
        expect(item.contains).toHaveBeenCalledWith('c');
      });
    });

    it('Given secondary to return manager THen must return expected response', function () {
      const instance = createInstance();

      const expectedResponse = {
        a: 9
      };
      instance.properties.primary = {
        contains: jasmine.createSpy('contains').and.returnValue(false)
      };
      instance.properties.secondaries = [{
        contains: jasmine.createSpy('contains').and.returnValue(false)
      }, {
        contains: jasmine.createSpy('contains').and.returnValue(true),
        get: jasmine.createSpy('get').and.returnValue(expectedResponse)
      }, {
        contains: jasmine.createSpy('contains').and.returnValue(false)
      }];

      expect(instance.get('c', 'j')).toBe(expectedResponse);
      expect(instance.properties.primary.contains).toHaveBeenCalledTimes(1);
      expect(instance.properties.primary.contains).toHaveBeenCalledWith('c');

      expect(instance.properties.secondaries[0].contains).toHaveBeenCalledTimes(1);
      expect(instance.properties.secondaries[0].contains).toHaveBeenCalledWith('c');

      expect(instance.properties.secondaries[1].contains).toHaveBeenCalledTimes(1);
      expect(instance.properties.secondaries[1].contains).toHaveBeenCalledWith('c');
      expect(instance.properties.secondaries[1].get).toHaveBeenCalledTimes(1);
      expect(instance.properties.secondaries[1].get).toHaveBeenCalledWith('c');

      expect(instance.properties.secondaries[2].contains).not.toHaveBeenCalled();
    });
  }); // #get

  describe('#add', function () {
    it('Given data Then must call primary.add', function () {
      const instance = createInstance();

      instance.properties.primary = {
        add: jasmine.createSpy('add')
      };

      const dependency = {
        a: 5
      };
      instance.add('a', dependency);
      expect(instance.properties.primary.add).toHaveBeenCalledTimes(1);
      expect(instance.properties.primary.add).toHaveBeenCalledWith('a', dependency);
    });
  }); // #add

  describe('#remove', function () {
    it('Given key Then must call primary.remove', function () {
      const instance = createInstance();

      instance.properties.primary = {
        remove: jasmine.createSpy('remove')
      };
      instance.remove('a');
      expect(instance.properties.primary.remove).toHaveBeenCalledTimes(1);
      expect(instance.properties.primary.remove).toHaveBeenCalledWith('a');
    });
  }); // #remove

  describe('#dependencyNames', function () {
    it('Then must return expected response', function () {
      const instance = createInstance();

      instance.properties.primary = {
        dependencyNames: jasmine.createSpy('dependencyNames').and.returnValue(['a', 'c'])
      };
      instance.properties.secondaries = [{
        dependencyNames: jasmine.createSpy('dependencyNames').and.returnValue(['a1', 'c1', 'a'])
      }, {
        dependencyNames: jasmine.createSpy('dependencyNames').and.returnValue(['a2', 'c2', 'c1'])
      }];

      expect(instance.dependencyNames()).toEqual(['a', 'a1', 'a2', 'c', 'c1', 'c2']);
      expect(instance.properties.secondaries[0].dependencyNames).toHaveBeenCalled();
      expect(instance.properties.secondaries[1].dependencyNames).toHaveBeenCalled();
    });
  }); // #dependencyNames
});
